# -*- coding: utf-8 -*-
"""
Created on Fri Apr 21 18:24:26 2017

@author: Timo
"""

import numpy as np
import numpy.linalg as LA
from .. import util
import math as m
import scipy.optimize as opt
import pyFAI
import pyFAI.azimuthalIntegrator
from pyFAI.ext import invert_geometry
import copy

from .HKLVlieg import crystalAngles_singleArray

def primBeamAngles(pos):
    [alpha,delta,gamma,omega,chi,phi] = pos
    gamma_p = np.arcsin( np.cos(alpha)*np.sin(gamma) + np.sin(alpha)*np.cos(delta)*np.cos(gamma) )
    delta_p = np.arcsin( (np.sin(delta)*np.cos(gamma))/np.cos(gamma_p) )
    return [alpha,delta_p,gamma_p,omega,chi,phi]

def vliegDiffracAngles(pos_p):
    [alpha,delta_p,gamma_p,omega,chi,phi] = pos_p
    gamma = np.arcsin( np.cos(alpha)*np.sin(gamma_p) - np.sin(alpha)*np.cos(delta_p)*np.cos(gamma_p) )
    delta = np.arcsin( (np.sin(delta_p)*np.cos(gamma_p))/np.cos(gamma) )
    return [alpha,delta,gamma,omega,chi,phi]



"""

Attention!!!!

There is something wierd with pyFAI

x,y seem to be inverted

"""
class Detector2D_SXRD(pyFAI.azimuthalIntegrator.AzimuthalIntegrator):
    
    def __init__(self,*args,**keyargs):
        super(Detector2D_SXRD,self).__init__(*args,**keyargs)
        self.setAzimuthalReference(0)
        self.setPolarization(0,0)
    
    #in rad!!!
    def setAzimuthalReference(self,deltaChi):
        self._deltaChi = copy.copy(deltaChi)
        self._cached_array = {}
        
    def setPolarization(self,angle,factor):
        self._polAxis = angle
        self._polFactor = factor
        self._cached_array = {}
        
    def primBeamAngles(self,shape=None):
        # gives angles in flat-panel reference frame 
        
        if self._cached_array.get("gamma_p") is None or self._cached_array.get("delta_p") is None:
            azimuth = self.chiArray(shape=shape) + self._deltaChi
            tth = self.array_from_unit(unit=pyFAI.units.TTH_RAD)
            gamma_p = np.arctan(np.cos(azimuth)*np.tan(tth))
            delta_p = np.arctan(np.sin(azimuth)*np.tan(tth))
            self._cached_array['gamma_p'] = gamma_p
            self._cached_array['delta_p'] = delta_p
        
        return self._cached_array['gamma_p'], self._cached_array['delta_p']
    
    #numpy array!!!
    def primBeamPoints(self,x,y):
        azimuth = self.chi(x,y) + self._deltaChi
        tth = self.tth(x,y)        
        gamma_p = np.arctan(np.cos(azimuth)*np.tan(tth))
        delta_p = np.arctan(np.sin(azimuth)*np.tan(tth))
        return gamma_p, delta_p
    
    def surfaceAngles(self,alpha_i,shape=None):
        if hasattr(self,'_alpha_i'):
            if self._alpha_i == alpha_i:
                if self._cached_array.get("gamma") is not None and self._cached_array.get("delta") is not None:
                    return self._cached_array.get("gamma"), self._cached_array.get("delta")
        self._alpha_i = alpha_i           
        gamma_p, delta_p = self.primBeamAngles(shape)
        gamma = np.arcsin( np.cos(alpha_i)*np.sin(gamma_p) - np.sin(alpha_i)*np.cos(delta_p)*np.cos(gamma_p) )
        delta = np.arcsin( (np.sin(delta_p)*np.cos(gamma_p))/np.cos(gamma) )
        self._cached_array["gamma"] = gamma
        self._cached_array["delta"] = delta
        return self._cached_array.get("gamma"), self._cached_array.get("delta")
    
    def crystalAngles(self,alpha_i,refraction_index,shape=None):
        if hasattr(self,'_n_refr'):
            if self._n_refr == refraction_index:
                if self._cached_array.get("gamma_cry") is not None and hasattr(self,'_alpha_i_ref'):
                    return self._cached_array.get("gamma_cry"), self._cached_array.get("delta"), self._alpha_i_ref
        self._n_refr = refraction_index
        gamma, delta = self.surfaceAngles(alpha_i,shape)
        self._cached_array["gamma_cry"] = crystalAngles_singleArray(gamma,self._n_refr)
        self._alpha_i_ref = crystalAngles_singleArray(self._alpha_i,self._n_refr)
        return self._cached_array.get("gamma_cry"), self._cached_array.get("delta"), self._alpha_i_ref
        
        
    def surfaceAnglesPoint(self,x,y,alpha_i):
        gamma_p, delta_p = self.primBeamPoints(x,y)
        gamma = np.arcsin( np.cos(alpha_i)*np.sin(gamma_p) - np.sin(alpha_i)*np.cos(delta_p)*np.cos(gamma_p) )
        delta = np.arcsin( (np.sin(delta_p)*np.cos(gamma_p))/np.cos(gamma) )
        return gamma, delta
    
    def crystalAnglesPoint(self,x,y,alpha_i,refraction_index):
        gamma, delta = self.surfaceAnglesPoint(x,y,alpha_i)
        gamma_cry = crystalAngles_singleArray(gamma,refraction_index)
        alpha_i_cry = crystalAngles_singleArray(alpha_i,refraction_index)
        return gamma_cry, delta, alpha_i_cry
    
    # for now numerical solution, only iterables!
    # in rad
    def pixelsPrimPoint(self,gamma_p,delta_p,shape=None):
        gamma_p_det, delta_p_det = self.primBeamAngles(shape)
        ig = invert_geometry.InvertGeometry(gamma_p_det,delta_p_det)
        xy = []
        for gam_p, del_p in zip(gamma_p,delta_p):
            xy.append(ig(gam_p,del_p))
        return xy
    
    def pixelsSurfacePoint(self,gamma,delta,alpha_i,shape=None):
        gamma_det, delta_det = self.surfaceAngles(alpha_i,shape)
        ig = invert_geometry.InvertGeometry(gamma_det,delta_det)
        xy = []
        for gam, delt in zip(gamma,delta):
            xy.append(ig(gam,delt))
        return xy
    
    
    def correctionArray(self,shape=None):
        if shape is None:
            shape = self.get_shape()
        if self._cached_array.get("corrarr") is not None:
            return self._cached_array.get("corrarr")
        else:
            self._cached_array["corrarr"] = self.solidAngleArray(shape)*self.polarization(shape=shape,factor=self._polFactor,axis_offset=self._polAxis)
        return self._cached_array.get("corrarr")
    

class DetectorCalibration():
    
    def __init__(self,E,calibrationCrystal,pixelsize):
        self._energy = E
        self._crystal = calibrationCrystal
        self._pixelsize = pixelsize

    def setE(self,E):
        self._energy = E

    # shapes: hkl = [[h1,k1,l1], [h2,k2,l2], ...] , xy = [[x1,y1], [x2,y2], ...]
    def calibrateFromReflections(self,hkl,xy):
        x_centralPixel = 730.
        xy = xy.T
        xyz = np.zeros((3,xy.shape[1]))
        xyz[:2] = xy
        
        twoTheta = self._crystal.get2ThetaFromHKL(hkl,self._energy)
        
        """
        p[0]: th_x, p[1]: th_y, p[2]: d_detector, p[3]: y_centralPixel
        """
        def Chi2(p):
            centralPixel = np.array([x_centralPixel,p[3],0.])
            _xyz = np.matrix((xyz.T - centralPixel).T)
            # uncomment to enable detector tilt correction:
            #_xyz = util.x_rotation(p[0])*util.y_rotation(p[1])*_xyz
            _xyz[2] = 0
            d = LA.norm(_xyz,axis=0)*self._pixelsize
            print(list((twoTheta - np.arctan(d/p[2])) / twoTheta ))
            return LA.norm((twoTheta - np.arctan(d/p[2]))/twoTheta)
        
        fitbounds = [(-0.1,0.1),(-0.1,0.1),(0.5,1.),(1400,1430)]
        res = opt.minimize(Chi2,np.array([0.0,0.0,0.72,1415.]),bounds=fitbounds,method='TNC')
        print(res)
        res = opt.minimize(Chi2,res.x)
        print(res)
        self.setCalibration([x_centralPixel,res.x[3]],res.x[2],[res.x[0],res.x[1]])
        self._rotz = None
        
    def setCalibration(self,centralPixel,distance,rotz=None):
        self._centralPixel = centralPixel
        self._distance = distance
        self._rotz = rotz
        
    def xyToDelGam(self,x,y):
        x = (x - self._centralPixel[0])*self._pixelsize
        y = (y - self._centralPixel[1])*-self._pixelsize
        #print(x)
        xy = np.vstack([x,y])
        if self._rotz is not None:
            mat = np.matrix([[np.cos(self._rotz),-np.sin(self._rotz)],
                             [np.sin(self._rotz),np.cos(self._rotz)]])
            #print(xy)
            xy = mat*xy
            xy = xy.A
            #print(xy)
            #raise NotImplementedError("Detector tilt is not implemented")
        
        delta = np.arctan(xy[0]/self._distance)
        gamma = np.arctan(xy[1]/self._distance)
        
        #print(delta)
        return delta , gamma
    
    def delGamToxy(self,delta,gamma):
        x = np.tan(delta)*self._distance
        y = np.tan(gamma)*self._distance
        
        x = (x/self._pixelsize) + self._centralPixel[0]
        y = -(y/self._pixelsize) + self._centralPixel[1]
        
        return [x,y]
    
    def xyToDelGam_corr(self,x,y):
        x = (x - self._centralPixel[0])*self._pixelsize
        y = (y - self._centralPixel[1])*-self._pixelsize
        #print(x)
        #xy = np.vstack([x,y])
        """
        if self._rotz is not None:
            mat = np.matrix([[np.cos(self._rotz),-np.sin(self._rotz)],
                             [np.sin(self._rotz),np.cos(self._rotz)]])
            #print(xy)
            xy = mat*xy
            xy = xy.A
            #print(xy)
            #raise NotImplementedError("Detector tilt is not implemented")
        """
        delta = np.arctan(x/self._distance)
        gamma = np.arctan(y/self._distance)
        
        deldelta = np.arctan((x+self._pixelsize)*(1/self._distance)) - delta
        delgamma = np.arctan((y+self._pixelsize)*(1/self._distance)) - gamma
        
        dd,dg = np.meshgrid(deldelta,delgamma)
                
        correction = dd*dg
        correction /= np.amax(correction)
        #print(delta)
        return delta , gamma, correction
    
    def xyToDelGam_grid(self,x,y):
        x = (np.array(x,dtype=np.float) - self._centralPixel[0])*self._pixelsize
        y = (np.array(y,dtype=np.float) - self._centralPixel[1])*-self._pixelsize
        #print(x)
            
        xx, yy = np.meshgrid(x,y)
        
        if self._rotz is not None:
            xx = np.cos(self._rotz)*xx - np.sin(self._rotz)*yy
            yy = np.sin(self._rotz)*xx + np.cos(self._rotz)*yy
        
        
        
        delta = np.arctan(xx*(1/self._distance))
        gamma = np.arctan(yy*(1/self._distance))
        
        deldelta = np.arctan((xx+self._pixelsize)*(1/self._distance)) - delta
        delgamma = np.arctan((yy+self._pixelsize)*(1/self._distance)) - gamma
        
        correction = deldelta*delgamma
        correction /= np.amax(correction)
        
        return delta , gamma, correction
    
    # **** Depricated  ****
    # Doesn't work good with just 4 points, but nice idea
    # P1 and P2 are one pair, P3, P4 the other, lower theta first
    def powderCalibration(self,P1,P2,P3,P4):
        hkl1 , pos1 = P1 
        hkl2 , pos2 = P2
        hkl3 , pos3 = P3
        hkl4 , pos4 = P4
        
        twoTheta1 = self._crystal.get2ThetaFromHKL(hkl1,self._energy)
        twoTheta2 = self._crystal.get2ThetaFromHKL(hkl2,self._energy)
        gamma = twoTheta2 -twoTheta1
        
        l1 = LA.norm(pos1 - pos2)*pixelsize
        l2 = LA.norm(pos3 - pos4)*pixelsize
        
        delta = util.solveTrigEquation(twoTheta1,twoTheta2,l1,l2)
        
        b1 = (l1/m.sin(gamma))*m.cos(delta - twoTheta2)
        
        A = (m.cos(delta - twoTheta1) / m.sin(twoTheta1) )**2 - 1.
        B = -2. * b1 *m.cos(delta - twoTheta1) * (1/m.tan(twoTheta1))
        print(b1)
        
        ld = - (B/(2*A)) + m.sqrt((B/(2*A))**2 - (b1**2)/A)
        print(m.sqrt((B/(2*A))**2 - (b1**2)/A))
        
        print("angle : %s, distance: %s pixel" % (np.rad2deg(delta),ld/pixelsize ))
        
        direction = (pos1 - pos2)/LA.norm(pos1 - pos2)
        print ("Central pixel: %s" % (direction*ld/pixelsize + pos1))
        #print("prev: %s" % LA.norm(pos1 - centralPixel))

"""
pt111 = Crystal([ 2.7748 , 2.7748 , 6.7969],[  90. ,  90. , 120.])

dc = DetectorCalibration(E,pt111,pixelsize)
#hkl = fullrs_0007_reflections[:3].T
#xy = fullrs_0007_reflections[3:].T
hkl = sixc0015_0015_reflections[:3].T
xy = sixc0015_0015_reflections[3:].T

dc.calibrateFromReflections(hkl,xy)
        

P1 = ([-1. , 0. , 1.], fullrs_0007_reflections[1][0])
P2 = ([-2. , 0. , 2.], fullrs_0007_reflections[2][0])
P3 = ([1. , 0. , -1.], fullrs_0007_reflections[1][3])
P4 = ([2. , 0. , -2.], fullrs_0007_reflections[2][3])

dc = DetectorCalibration(E,pt111)
dc.powderCalibration(P1,P2,P3,P4)

P1 = ([1. , 0. , 1.], fullrs_0007_reflections[1][1])
P2 = ([2. , 0. , 2.], fullrs_0007_reflections[2][1])
P3 = ([-1. , 0. , -1.], fullrs_0007_reflections[1][2])
P4 = ([-2. , 0. , -2.], fullrs_0007_reflections[2][2])

dc.powderCalibration(P1,P2,P3,P4)
"""
        
        
        
        
        
        
        
        
    
